function model = trainModel( fileStr, params )
% Trains foreground and background region model
%   Computes foreground and background histograms from manually drawn
%   labels.
% 
%   (c) 2015 Robert Bensch, Computer Science Department, University of Freiburg, Germany
%
    
    %
    % input parameters
    %
    doPreSmoothing = params.doPreSmoothing;
    preSmoothSigma = params.preSmoothSigma;
    doBackgroundSubstr = params.doBackgroundSubstr;
    bgrSubstrSigma = params.bgrSubstrSigma;
    nBins = params.nBins;
    
    %
    % pre-compute smoothing kernels
    %
    hsize = ceil(2*(3*preSmoothSigma));
    if (mod(hsize,2)==0)
        hsize = hsize + 1;
    end
    kernel_preSmoothing = fspecial('gaussian', hsize, preSmoothSigma);

    hsize = ceil(2*(3*bgrSubstrSigma));
    if (mod(hsize,2)==0)
        hsize = hsize + 1;
    end
    kernel_bgrSubstr = fspecial('gaussian', hsize, bgrSubstrSigma);
    
    %
    % get input files and path
    %
    d = dir( fileStr);
    nFrames = length(d);
    idx = strfind(fileStr, '/');
    path = '';
    if (~isempty(idx))
        path = fileStr(1:idx(end));
    end
        
    %
    % initialize histograms
    %
    hist_fgr = zeros(1,nBins);
    hist_bgr = zeros(1,nBins);
    bin_step = 1/nBins;
    bin_range = [0:bin_step:1];
    hist_binCenters = bin_range(1:end-1) + bin_step/2;
    
    %
    % iterate over input images
    %
    fprintf('\nprocessing %d file(s) in total.\n', nFrames);
    for frameIdx=1:nFrames
    
        fn = d(frameIdx).name;
 
        fprintf('\tprocessing %s\n', fn);
        
        % read image
        im = im2double(imread([path fn]));
        %figure; imshow(im);
        
        % pre-smoothing
        if (doPreSmoothing)
            im = imfilter( im, kernel_preSmoothing, 'symmetric');    
            figure; imshow(im);
        end
        
        % backgorund substraction
        if (doBackgroundSubstr)
           im_bgr = imfilter( im, kernel_bgrSubstr, 'symmetric');  
           im = 0.5 * ((im - im_bgr) + 1.0);    % > [0,1]
           figure; imshow(im);
        end
       
        % create manual masks
        mask = createMasks( im, params );

%         im_overlay = cat(3, im+0.2*double(mask==1), im+0.2*double(mask==2), im);
%         im_overlay = im_overlay / max(im_overlay(:));
%         figure; imshow(im_overlay);
%         title('Foreground and background labels.');

        % compute histograms
        [hist_fgr_incr, ~] = hists( im(mask==2), hist_binCenters);
        [hist_bgr_incr, ~] = hists( im(mask==1), hist_binCenters);
     
        % accumulate histograms
        hist_fgr = hist_fgr + hist_fgr_incr;
        hist_bgr = hist_bgr + hist_bgr_incr;
                
        %close all;
        
    end
      
    %
    % normalize histograms
    %
    if (sum(hist_fgr)>0)
      hist_fgr = hist_fgr./sum(hist_fgr);
    end
    if (sum(hist_bgr)>0)
      hist_bgr = hist_bgr./sum(hist_bgr);
    end
    
    %
    % plot histograms
    %
    figure;
    hArray = bar(hist_binCenters', [hist_fgr; hist_bgr]');
    title('Foreground/background histograms');
    grid on;
    set(hArray(1), 'FaceColor', 'green', 'EdgeColor', 'green');
    set(hArray(2), 'FaceColor', 'red', 'EdgeColor', 'red');
    ylim([0 1]);
    xlabel('Intensity (normalized)'); ylabel('Relative frequency');
    legend('Foreground histogram','Background histogram');   

    
    %
    % output model
    %
    model = struct;
    model.hist_fgr = hist_fgr;
    model.hist_bgr = hist_bgr;
    model.hist_binCenters = hist_binCenters;
    params.kernel_preSmoothing = kernel_preSmoothing;
    params.kernel_bgrSubstr = kernel_bgrSubstr;
    model.params = params;
    
end

